////////////////////////////////////////////////////////////////////////////////
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.parser;

import net.objectlab.qalab.util.QALabTags;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * As part of the up/down movers task, this class contains info about a given
 * file.
 * 
 * @author Paramjit Rehinsi
 * @version $Revision: 1.17 $
 */
public class FileStats {
    private static final int ONE_DAY = 24;

    /** the date cut-off, do not include stats if before that date. */
    private Date startDateCutoff;

    /** the end date cut-off, do not include stats if after that date. */
    private Date endDateCutoff = null;

    /** buffer for the UP statistics (increased number of violations). */
    private StringBuffer upBuf = new StringBuffer();

    /** buffer for the DOWN statistics (decreased number of violations). */
    private StringBuffer downBuf = new StringBuffer();

    /** file name (typically a .java file). */
    private String filename;

    /** path to file. */
    private String pathToFile;

    /** type surveyed (checkstyle|findbugs|simian|pmd). */
    private Map typesSurveyed;
    
    /** offset in hours to determine if the stat has fallen to 0 over the last n hours
     * from the endDateCutoff.
     */
    private int offsetForLastRun = ONE_DAY;

    /**
     * @param types
     *            the types surveyed.
     * @param fileSurveyed
     *            the file name surveyed.
     * @param cutoff
     *            earliest date to take into account.
     */
    public FileStats(final Map types, final String fileSurveyed, final String pathToFile, final Date cutoff) {
        this.typesSurveyed = types;
        this.filename = fileSurveyed;
        this.pathToFile = pathToFile;
        this.startDateCutoff = new Date(cutoff.getTime());
    }

    /**
     * @return Return filename of this object.
     */
    public final String getFileName() {
        return filename;
    }

    /**
     * @return Return the XML entry for UP movers (increased violations).
     */
    public final String getUpXmlResult() {
        return upBuf.toString();
    }

    /**
     * @return Return the XML entry for DOWN movers (decreased violations).
     */
    public final String getDownXmlResult() {
        return downBuf.toString();
    }

    /**
     * Construct the XML buffers.
     */
    public final void constructXML() {
        Set keys = (Set) typesSurveyed.keySet();
        Iterator i = keys.iterator();

        while (i.hasNext()) {
            Stack s = (Stack) typesSurveyed.get((String) i.next());

            processFilesResultBeans(s);
        }

        if (upBuf.length() > 0) {
            upBuf.insert(0, "  <file name=\"" + getFileName() + "\" path=\"" + getPathToFile() + "\">" + QALabTags.LINE_END);
            upBuf.append("  </file>").append(QALabTags.LINE_END);
        }

        if (downBuf.length() > 0) {
            downBuf.insert(0, "  <file name=\"" + getFileName() + "\" path=\"" + getPathToFile() + "\">" + QALabTags.LINE_END);
            downBuf.append("  </file>").append(QALabTags.LINE_END);
        }
    }

    /**
     * Build the XML string depending on whether differences have occured since
     * the cutoff.
     * 
     * @param stack
     *            each result set in a stack from cutoff.
     */
    private void processFilesResultBeans(final Stack stack) {
        if (stack.empty()) {
            return;
        }

        // get a result from stack.
        SingleStat currentResult = (SingleStat) stack.pop();

        Iterator it = stack.iterator();
//        boolean onlyOneResult = !it.hasNext();

        // set the date of the result.
        Calendar currentday = Calendar.getInstance();
        currentday.setTime(getEndDateCutoff());
        currentday.add(Calendar.HOUR, -offsetForLastRun);

        // new result -> gone UP!
        // BUT only if the result is not BEFORE 1 day from the end result
        // (otherwise we assume that the has been another run which reduced it
        // to 0, therefore it would not have gone up.
//        if (onlyOneResult && fallInWindow(currentResult)
        /** && !currentResult.getDate().before(currentday.getTime()) */
//        ) {
//            appendStats(upBuf, null, currentResult);
//        }

        // if the result set is after the cutoff and less than end-1 then
        // it means that we do NOT have a result for this guy on end
        if (/*onlyOneResult && */!currentResult.getDate().before(startDateCutoff)
                && currentResult.getDate().before(currentday.getTime())) {
            appendStats(downBuf, currentResult, null);
        }

        checkPreviousResults(stack, currentResult, it);
    }

    /**
     * @param stack
     * @param currentResult
     * @param it
     */
    private void checkPreviousResults(final Stack stack, SingleStat currentResult, Iterator it) {
        SingleStat previousResult;
        Calendar oneDayBeforeEndCutoff = Calendar.getInstance();
        oneDayBeforeEndCutoff.setTime(endDateCutoff);
        oneDayBeforeEndCutoff.add(Calendar.HOUR, -offsetForLastRun);

        // iterate through the previous results.
        while (it.hasNext()) {
            previousResult = (SingleStat) stack.pop();

            int previousCount = previousResult.getStatValue();
            final int currentCount = currentResult.getStatValue();

            // if (previousResult.getDate().before(startDateCutoff)) {
            // previousCount = 0; // assume 0.
            // }

            final int diff = currentCount - previousCount;

            if (diff > 0) {
                handleUp(currentResult, previousResult);
            } else if (diff < 0) {
                handleDown(currentResult, previousResult);
            } else if (diff == 0 && fallInWindow(currentResult)
                    && !currentResult.getDate().before(oneDayBeforeEndCutoff.getTime())
                    && !previousResult.getDate().after(startDateCutoff)) {
                // this is the case where we may have had the same number of
                // violation outside the window in the past but this should
                // count as a NEW violation as it has come back.
                appendStats(upBuf, null, currentResult);
            }

            currentResult = previousResult;
        }
        // if the last resul (earliest) falls in the window, we can assume that
        // it should be part of the UP category. 
        if (fallInWindow(currentResult)) {
            appendStats(upBuf, null, currentResult);
        }
    }

    /**
     * @param currentResult
     * @param previousResult
     */
    private void handleDown(SingleStat currentResult, SingleStat previousResult) {
        if (fallInWindow(currentResult)) {
            // there are LESS violations this time round.
            appendStats(downBuf, previousResult, currentResult);
        }
    }

    /**
     * @param currentResult
     * @param previousResult
     */
    private void handleUp(SingleStat currentResult, SingleStat previousResult) {
        if (fallInWindow(currentResult)) {
            // there are MORE violations this time round.
            appendStats(upBuf, previousResult, currentResult);
        }

    }

    /**
     * return true if the stat falls in the time window.
     * 
     * @param stat
     * @return
     */
    private boolean fallInWindow(SingleStat stat) {
        return !stat.getDate().before(startDateCutoff) && !stat.getDate().after(endDateCutoff);
    }

    /**
     * Append statistics to up or down buffer.
     * 
     * @param buffer
     *            the up or down buffer
     * @param previousResult
     *            can be null (ie 0)
     * @param currentResult
     *            can be null (ie 0)
     */
    private void appendStats(final StringBuffer buffer, final SingleStat previousResult, final SingleStat currentResult) {
        buffer.append("   <diff previousrun=\"");

        if (previousResult != null) {
            writeDate(buffer, previousResult);
        } else {
            buffer.append("-");
        }

        buffer.append("\" currentrun=\"");

        if (currentResult != null) {
            writeDate(buffer, currentResult);
            buffer.append("\" type=\"").append(currentResult.getType());
        } else {
            buffer.append("-");
            buffer.append("\" type=\"").append(previousResult.getType());
        }

        buffer.append("\" previouserrors=\"");

        int previousCount = 0;

        if (previousResult != null) {
            previousCount = previousResult.getStatValue();
        }

        int currentCount = 0;

        if (currentResult != null) {
            currentCount = currentResult.getStatValue();
        }

        buffer.append(previousCount);
        buffer.append("\" currenterrors=\"");
        buffer.append(currentCount);
        buffer.append("\" diff=\"");
        buffer.append(currentCount - previousCount);
        buffer.append("\"/>").append(QALabTags.LINE_END);
    }

    /**
     * @param buffer
     * @param currentResult
     */
    private void writeDate(final StringBuffer buffer, final SingleStat currentResult) {
        Date date = currentResult.getDate();
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);

        if (0 == cal.get(Calendar.SECOND) && 0 == cal.get(Calendar.MINUTE) && 0 == cal.get(Calendar.HOUR)) {
            buffer.append(QALabTags.DEFAULT_DATE_FORMAT.format(currentResult.getDate()));
        } else {
            buffer.append(QALabTags.DEFAULT_DATETIME_FORMAT.format(currentResult.getDate()));
        }
    }

    /**
     * @param endDate
     *            The endDateCutoff to set.
     */
    public final void setEndDateCutoff(final Date endDate) {
        this.endDateCutoff = new Date(endDate.getTime());
    }

    /**
     * @return Returns the endDateCutoff.
     */
    public final Date getEndDateCutoff() {
        return new Date(endDateCutoff.getTime());
    }

    /**
     * @return Returns the pathToFile.
     */
    public final String getPathToFile() {
        return pathToFile;
    }

    /**
     * @param pathToFile
     *            The pathToFile to set.
     */
    public final void setPathToFile(String pathToFile) {
        this.pathToFile = pathToFile;
    }

    public int getOffsetForLastRun() {
        return offsetForLastRun;
    }

    public void setOffsetForLastRun(int offsetForLastRun) {
        this.offsetForLastRun = offsetForLastRun;
    }
}
